# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import pytest

from capella_ros_tools.data_model import (
    ConstantDef,
    EnumDef,
    FieldDef,
    MessageDef,
    Range,
    TypeDef,
)


@pytest.fixture
def sample_class_def():
    return MessageDef(
        name="SampleClass",
        fields=[
            FieldDef(
                type=TypeDef("uint8", Range("0", "10"), None),
                name="sample_field1",
                description="This block comment is added to the "
                "property description of sample_field1. "
                "This block comment is also added to the "
                "property description of sample_field1. ",
            ),
            FieldDef(
                type=TypeDef("SampleClassEnum", Range("0", "*"), "package2"),
                name="sample_field2",
                description="This block comment is added to the property "
                "descriptions of sample_field2 and sample_field3. ",
            ),
            FieldDef(
                TypeDef("uint8", Range("3", "3"), None),
                name="sample_field3",
                description="This block comment is added to the property "
                "descriptions of sample_field2 and sample_field3. ",
            ),
            FieldDef(
                type=TypeDef("SampleEnum", Range("1", "1"), "SampleEnum"),
                name="sample_field4",
                description="This block comment is added to the property "
                "descriptions of sample_field4 and sample_field5. "
                "Fields in SampleClass can reference "
                "enums in other files. "
                "The property sample_field4 "
                "is of type SampleEnum. "
                "cf. SampleEnum ",
            ),
            FieldDef(
                type=TypeDef("SampleEnumValue", Range("1", "1"), "SampleEnum"),
                name="sample_field5",
                description="This block comment is added to the property "
                "descriptions of sample_field4 and sample_field5. "
                "This inline comment "
                "is added to the "
                "property description of "
                "sample_field5. "
                "The property sample_field5 "
                "is of type SampleEnumValue. "
                "cf. SampleEnum, SAMPLE_ENUM_VALUE_XXX ",
            ),
        ],
        enums=[],
        description="SampleClass.msg "
        "The first comment block at the top of the file "
        "is added to the class description of SampleClass. ",
    )


@pytest.fixture
def sample_enum_def():
    return MessageDef(
        name="SampleEnum",
        fields=[],
        enums=[
            EnumDef(
                name="SampleEnumValue",
                literals=[
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="RED",
                        value="0",
                        description="",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="BLUE",
                        value="1",
                        description="This inline comment "
                        "is added to the "
                        "enum literal "
                        "description of BLUE. ",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="YELLOW",
                        value="2",
                        description="This block comment is added to the "
                        "enum literal descriptions of YELLOW and GREEN. ",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="GREEN",
                        value="3",
                        description="This block comment is added to the "
                        "enum literal descriptions of YELLOW and GREEN. ",
                    ),
                ],
                description="SampleEnum.msg "
                "This block comment is added to the "
                "enum description of SampleEnumValue. ",
            ),
            EnumDef(
                name="SampleEnum",
                literals=[
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="OK",
                        value="0",
                        description="",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="WARN",
                        value="1",
                        description="",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="ERROR",
                        value="2",
                        description="",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="STALE",
                        value="3",
                        description="",
                    ),
                ],
                description="This block comment is added to the "
                "enum description of SampleEnum. "
                "In a file, there can only be one or no enum "
                "whose literal names do not share a common prefix. ",
            ),
        ],
        description="",
    )


@pytest.fixture
def sample_class_enum_def():
    return MessageDef(
        name="SampleClassEnum",
        fields=[
            FieldDef(
                type=TypeDef(
                    "SampleClassEnumStatus",
                    Range("1", "1"),
                    "SampleClassEnum",
                ),
                name="status",
                description="The property status is of type "
                "SampleClassEnumStatus. ",
            ),
            FieldDef(
                type=TypeDef("Color", Range("1", "1"), "SampleClassEnum"),
                name="color",
                description="The property color is of type Color. ",
            ),
            FieldDef(
                type=TypeDef("uint8", Range("1", "1"), None),
                name="field",
                description="",
            ),
        ],
        enums=[
            EnumDef(
                name="SampleClassEnumStatus",
                literals=[
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="OK",
                        value="0",
                        description="",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="WARN",
                        value="1",
                        description="",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="ERROR",
                        value="2",
                        description="",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="STALE",
                        value="3",
                        description="",
                    ),
                ],
                description="This block comment is added to the "
                "enum description of SampleClassEnumStatus. ",
            ),
            EnumDef(
                name="Color",
                literals=[
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="RED",
                        value="0",
                        description="",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="BLUE",
                        value="1",
                        description="",
                    ),
                    ConstantDef(
                        type=TypeDef("uint8", Range("1", "1"), None),
                        name="YELLOW",
                        value="2",
                        description="",
                    ),
                ],
                description="This block comment is added to the "
                "enum description of Color. ",
            ),
        ],
        description="SampleClassEnum.msg "
        "Properties in SampleClassEnum can reference "
        "enums in the same file. ",
    )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Tool for importing ROS messages to a Capella data package."""

import os
import pathlib
import re
import typing as t

from capellambse import decl, filehandler, helpers

from capella_ros_tools import data_model

from . import logger

ROS2_INTERFACES = {
    "common_interfaces": "git+https://github.com/ros2/common_interfaces",
    "rcl_interfaces": "git+https://github.com/ros2/rcl_interfaces",
    "unique_identifier_msgs": (
        "git+https://github.com/ros2/unique_identifier_msgs"
    ),
}


class Importer:
    """Class for importing ROS messages to a Capella data package."""

    def __init__(
        self,
        msg_path: str,
        no_deps: bool,
        license_header_path: pathlib.Path | None = None,
        msg_description_regex: str | None = None,
    ):
        self.messages = data_model.MessagePkgDef("root", [], [])
        self._promise_ids: dict[str, None] = {}
        self._promise_id_refs: dict[str, None] = {}
        self._license_header = None
        if license_header_path is not None:
            self._license_header = license_header_path.read_text("utf-8")

        self._add_packages("ros_msgs", msg_path, msg_description_regex)
        if no_deps:
            return

        for interface_name, interface_url in ROS2_INTERFACES.items():
            self._add_packages(interface_name, interface_url)

    def _add_packages(
        self, name: str, path: str, msg_description_regex: str | None = None
    ) -> None:
        root = filehandler.get_filehandler(path).rootdir
        msg_description_pattern = None
        if msg_description_regex is not None:
            msg_description_pattern = re.compile(
                msg_description_regex, re.MULTILINE
            )

        for dir in sorted(root.rglob("msg"), key=os.fspath):
            pkg_name = dir.parent.name or name
            pkg_def = data_model.MessagePkgDef.from_msg_folder(
                pkg_name, dir, self._license_header, msg_description_pattern
            )
            self.messages.packages.append(pkg_def)
            logger.info("Loaded package %s from %s", pkg_name, dir)

    def _convert_datatype(self, promise_id: str) -> dict[str, t.Any]:
        name = promise_id.split(".", 1)[-1]
        if any(t in name for t in ["char", "str"]):
            _type = "StringType"
        elif any(t in name for t in ["bool", "byte"]):
            _type = "BooleanType"
        else:
            _type = "NumericType"
        yml = {
            "promise_id": promise_id,
            "find": {
                "name": name,
                "_type": _type,
            },
        }
        return yml

    def _convert_package(
        self,
        pkg_def: data_model.MessagePkgDef,
    ) -> dict[str, t.Any]:
        classes = []
        enums = []
        packages = []
        associations = []

        for msg_def in pkg_def.messages:
            if msg_def.fields:
                cls_yml, cls_associations = self._convert_class(
                    pkg_def.name, msg_def
                )
                classes.append(cls_yml)
                associations.extend(cls_associations)
            for enum_def in msg_def.enums:
                enums.append(self._convert_enum(msg_def.name, enum_def))

        for new_pkg in pkg_def.packages:
            new_yml = {
                "find": {
                    "name": new_pkg.name,
                },
            } | self._convert_package(new_pkg)
            packages.append(new_yml)

        sync = {}
        if classes:
            sync["classes"] = classes
        if enums:
            sync["enumerations"] = enums
        if packages:
            sync["packages"] = packages
        if associations:
            sync["owned_associations"] = associations

        yml = {}
        if sync:
            yml["sync"] = sync

        return yml

    def _convert_class(
        self, pkg_name: str, msg_def: data_model.MessageDef
    ) -> tuple[dict[str, t.Any], list[dict[str, t.Any]]]:
        promise_id = f"{pkg_name}.{msg_def.name}"
        self._promise_ids[promise_id] = None
        props = []
        associations = []
        for field_def in msg_def.fields:
            prop_promise_id = f"{promise_id}.{field_def.name}"
            promise_ref = (
                f"{field_def.type.package or pkg_name}.{field_def.type.name}"
            )
            self._promise_id_refs[promise_ref] = None
            prop_yml = {
                "promise_id": prop_promise_id,
                "find": {
                    "name": field_def.name,
                },
                "set": {
                    "type": decl.Promise(promise_ref),
                    "kind": "COMPOSITION",
                    "description": field_def.description,
                    "min_card": decl.NewObject(
                        "LiteralNumericValue", value=field_def.type.card.min
                    ),
                    "max_card": decl.NewObject(
                        "LiteralNumericValue", value=field_def.type.card.max
                    ),
                },
            }
            props.append(prop_yml)

            associations.append(
                {
                    "find": {
                        "navigable_members": [decl.Promise(prop_promise_id)],
                    },
                    "sync": {
                        "members": [
                            {
                                "find": {
                                    "type": decl.Promise(promise_id),
                                },
                                "set": {
                                    "_type": "Property",
                                    "kind": "ASSOCIATION",
                                    "min_card": decl.NewObject(
                                        "LiteralNumericValue", value="1"
                                    ),
                                    "max_card": decl.NewObject(
                                        "LiteralNumericValue", value="1"
                                    ),
                                },
                            }
                        ],
                    },
                }
            )

        yml = {
            "promise_id": promise_id,
            "find": {
                "name": msg_def.name,
            },
            "set": {
                "description": msg_def.description,
            },
            "sync": {
                "properties": props,
            },
        }
        return yml, associations

    def _convert_enum(
        self, pkg_name: str, enum_def: data_model.EnumDef
    ) -> dict[str, t.Any]:
        promise_id = f"{pkg_name}.{enum_def.name}"
        self._promise_ids[promise_id] = None
        literals = []
        for literal in enum_def.literals:
            literal_yml = {
                "find": {
                    "name": literal.name,
                },
                "set": {
                    "description": literal.description,
                    "value": decl.NewObject(
                        "LiteralNumericValue", value=literal.value
                    ),
                },
            }
            literals.append(literal_yml)
        yml = {
            "promise_id": promise_id,
            "find": {
                "name": enum_def.name,
            },
            "set": {
                "description": enum_def.description,
            },
            "sync": {
                "literals": literals,
            },
        }

        return yml

    def to_yaml(
        self,
        root_uuid: str,
        types_parent_uuid: str = "",
        types_uuid: str = "",
    ) -> str:
        """Import ROS messages into a Capella data package."""
        logger.info("Generating decl YAML")
        instructions = [
            {"parent": decl.UUIDReference(helpers.UUIDString(root_uuid))}
            | self._convert_package(self.messages),
        ]
        needed_types = [
            p for p in self._promise_id_refs if p not in self._promise_ids
        ]
        if not needed_types:
            return decl.dump(instructions)

        datatypes = [
            self._convert_datatype(promise_id) for promise_id in needed_types
        ]
        if types_uuid:
            instructions.append(
                {
                    "parent": decl.UUIDReference(
                        helpers.UUIDString(types_uuid)
                    ),
                    "sync": {"datatypes": datatypes},
                }
            )
        elif types_parent_uuid:
            instructions.append(
                {
                    "parent": decl.UUIDReference(
                        helpers.UUIDString(types_parent_uuid)
                    ),
                    "sync": {
                        "packages": [
                            {
                                "find": {"name": "Data Types"},
                                "sync": {"datatypes": datatypes},
                            }
                        ],
                    },
                }
            )
        else:
            raise ValueError(
                "Either types_parent_uuid or types_uuid must be provided"
            )
        return decl.dump(instructions)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import pathlib
import re

import pytest
from capellambse import decl, helpers

from capella_ros_tools.data_model import (
    ConstantDef,
    EnumDef,
    FieldDef,
    MessageDef,
    Range,
    TypeDef,
)
from capella_ros_tools.importer import Importer

# pylint: disable=redefined-outer-name

PATH = pathlib.Path(__file__).parent

SAMPLE_PACKAGE_PATH = PATH.joinpath("data/data_model/example_msgs")
CUSTOM_LICENSE_PACKAGE_PATH = PATH.joinpath(
    "data/data_model/custom_license_msgs"
)
DESCRIPTION_REGEX_PACKAGE_PATH = PATH.joinpath(
    "data/data_model/description_regex_msgs"
)
SAMPLE_PACKAGE_YAML = PATH.joinpath("data/data_model/example_msgs.yaml")
DUMMY_PATH = PATH.joinpath("data/empty_project_60")
CUSTOM_LICENSE_PATH = PATH.joinpath(
    "data/data_model/custom_license_header.txt"
)
EXPECTED_DESCRIPTION_SAMPLE_CLASS_ENUM = (
    "SampleClassEnum.msg "
    "Properties in SampleClassEnum can reference enums in the same file. "
)

EXPECTED_DESCRIPTION_REGEX = (
    "Message type for providing the made decision. "
    "An additional description line. <br>Expect linebreak "
)

ROOT = helpers.UUIDString("00000000-0000-0000-0000-000000000000")
SA_ROOT = helpers.UUIDString("00000000-0000-0000-0000-000000000001")


@pytest.fixture
def importer() -> Importer:
    return Importer(DUMMY_PATH.as_posix(), True)


def test_convert_datatype(importer: Importer):
    promise_id = "std_msgs.uint8"
    expected = {
        "promise_id": "std_msgs.uint8",
        "find": {
            "name": "uint8",
            "_type": "NumericType",
        },
    }

    actual = importer._convert_datatype(promise_id)

    assert decl.dump([actual]) == decl.dump([expected])


def test_convert_enum(importer: Importer):
    enum_def = EnumDef(
        name="MyEnum",
        description="An example enum",
        literals=[
            ConstantDef(
                type=TypeDef("uint8", Range("1", "1")),
                name="LITERAL_A",
                value="0",
                description="Literal A",
            ),
            ConstantDef(
                type=TypeDef("uint8", Range("1", "1")),
                name="LITERAL_B",
                value="1",
                description="Literal B",
            ),
        ],
    )
    expected = {
        "promise_id": "MyMessage.MyEnum",
        "find": {
            "name": "MyEnum",
        },
        "set": {
            "description": "An example enum",
        },
        "sync": {
            "literals": [
                {
                    "find": {
                        "name": "LITERAL_A",
                    },
                    "set": {
                        "description": "Literal A",
                        "value": decl.NewObject(
                            "LiteralNumericValue", value="0"
                        ),
                    },
                },
                {
                    "find": {
                        "name": "LITERAL_B",
                    },
                    "set": {
                        "description": "Literal B",
                        "value": decl.NewObject(
                            "LiteralNumericValue", value="1"
                        ),
                    },
                },
            ],
        },
    }

    actual = importer._convert_enum("MyMessage", enum_def)

    assert decl.dump([actual]) == decl.dump([expected])
    assert "MyMessage.MyEnum" in importer._promise_ids


def test_convert_class(importer: Importer):
    class_def = MessageDef(
        name="MyMessage",
        description="An example message",
        fields=[
            FieldDef(
                name="field",
                type=TypeDef("uint8", Range("1", "1")),
                description="Field",
            ),
        ],
        enums=[],
    )
    expected = {
        "promise_id": "my_package.MyMessage",
        "find": {
            "name": "MyMessage",
        },
        "set": {
            "description": "An example message",
        },
        "sync": {
            "properties": [
                {
                    "promise_id": "my_package.MyMessage.field",
                    "find": {
                        "name": "field",
                    },
                    "set": {
                        "type": decl.Promise("my_package.uint8"),
                        "kind": "COMPOSITION",
                        "description": "Field",
                        "min_card": decl.NewObject(
                            "LiteralNumericValue", value="1"
                        ),
                        "max_card": decl.NewObject(
                            "LiteralNumericValue", value="1"
                        ),
                    },
                },
            ],
        },
    }

    actual, _ = importer._convert_class("my_package", class_def)

    assert decl.dump([actual]) == decl.dump([expected])
    assert "my_package.MyMessage" in importer._promise_ids
    assert "my_package.uint8" in importer._promise_id_refs


def test_convert_class_with_ref(importer: Importer):
    pkg_name = "my_package"
    msg_def = MessageDef(
        name="MyMessage",
        description="An example message",
        fields=[
            FieldDef(
                name="field",
                type=TypeDef("uint8", Range("1", "1"), "std_msgs"),
                description="Field",
            ),
        ],
        enums=[],
    )
    expected = {
        "promise_id": "my_package.MyMessage",
        "find": {
            "name": "MyMessage",
        },
        "set": {
            "description": "An example message",
        },
        "sync": {
            "properties": [
                {
                    "promise_id": "my_package.MyMessage.field",
                    "find": {
                        "name": "field",
                    },
                    "set": {
                        "type": decl.Promise("std_msgs.uint8"),
                        "kind": "COMPOSITION",
                        "description": "Field",
                        "min_card": decl.NewObject(
                            "LiteralNumericValue", value="1"
                        ),
                        "max_card": decl.NewObject(
                            "LiteralNumericValue", value="1"
                        ),
                    },
                },
            ],
        },
    }

    actual, _ = importer._convert_class(pkg_name, msg_def)

    assert decl.dump([actual]) == decl.dump([expected])
    assert "my_package.MyMessage" in importer._promise_ids
    assert "std_msgs.uint8" in importer._promise_id_refs


def test_convert_package():
    expected = decl.dump(decl.load(SAMPLE_PACKAGE_YAML))

    actual = Importer(SAMPLE_PACKAGE_PATH.as_posix(), True).to_yaml(
        ROOT, SA_ROOT
    )

    assert actual == expected


def test_custom_license_header():
    importer = Importer(
        CUSTOM_LICENSE_PACKAGE_PATH.as_posix(), True, CUSTOM_LICENSE_PATH
    )

    assert (
        importer.messages.packages[0].messages[0].description
        == EXPECTED_DESCRIPTION_SAMPLE_CLASS_ENUM
    )


def test_description_regex():
    importer = Importer(
        DESCRIPTION_REGEX_PACKAGE_PATH.as_posix(),
        True,
        msg_description_regex=r"^Description:\s*([\s\S]*)",
    )

    assert (
        importer.messages.packages[0].messages[0].description
        == EXPECTED_DESCRIPTION_REGEX
    )
